import time
import struct
import json
import logging
from . import mainnet_chain
from .setting import (MAX_NONCE,
    DEFAULT_TX_FEE,
    DEFAULT_FEE_SCALE,
    DEFAULT_LEASE_FEE,
    DEFAULT_CANCEL_LEASE_FEE,
    PAYMENT_TX_TYPE,
    LEASE_TX_TYPE,
    LEASE_CANCEL_TX_TYPE,
    CHECK_FEE_SCALE,
    MAX_ATTACHMENT_SIZE)
from .error import (InvalidAddressException,
    InvalidParameterException,
    InsufficientBalanceException,
    MissingPrivateKeyException,
    NetworkException,
    throw_error)
from .crypto import *
from ..wallet_cipher import WalletCipher

class Account(object):

    def __init__(self, chain=mainnet_chain(), address='', privateKey='', seed='', alias='', nonce=0):
        self.chain = chain
        self.wrapper = chain.apiWrapper
        self.publicKey = ''
        if nonce < 0 or nonce > MAX_NONCE:
            raise InvalidParameterException(f'Nonce must be between 0 and {MAX_NONCE}')

        if seed:
            self._generate(seed=seed, nonce=nonce)
        elif privateKey:
            self._generate(privateKey=privateKey)
        else:
            raise InvalidParameterException('Seed or private key not provided for account')

    def __str__(self):
        if not self.address:
            raise InvalidAddressException('No address')

        result = 'address = %s\npublicKey = %s\nprivateKey = %s\nseed = %s\nnonce = %d' % \
            (self.address, self.publicKey, self.privateKey, self.seed, self.nonce)

        try:
            balance = self.balance()
            result += f'\nbalance: {balance}'
        except NetworkException:
            logging.error('Failed to get balance')
        return result

    __repr__ = __str__

    def _generate(self, seed='', privateKey='', nonce=0):
        self.seed = seed
        self.nonce = nonce
        if seed:
            accInfo = WalletCipher.generate_account_info(seed=seed, net=self.chain.chainId, nonce=nonce)
            self.address = accInfo['addr']
            self.privateKey = bytes2str(base58.b58encode(accInfo['priv']))
            self.publicKey = bytes2str(base58.b58encode(accInfo['pub']))
        elif privateKey:
            privKey = base58.b58decode(privateKey)
            pubKey = curve.generatePublicKey(privKey)
            self.address = self.chain.public_key_to_address(pubKey).decode('utf-8')
            self.privateKey = privateKey
            self.publicKey = bytes2str(base58.b58encode(pubKey))

    def balance(self, confirmations=0):
        try:
            confirmationStr = '' if confirmations == 0 else f'/{confirmations}'
            resp = self.wrapper.request(f'/addresses/balance/{str(self.address)}{confirmationStr}')
            return resp['balance']
        except Exception as ex:
            msg = f'Failed to get balance. ({ex})'
            logging.error(msg)
            return 0

    def balance_detail(self):
        try:
            resp = self.wrapper.request(f'/addresses/balance/details/{str(self.address)}')
            return resp
        except Exception as ex:
            msg = f'Failed to get balance detail. ({ex})'
            logging.error(msg)
            return None

    def send_payment(self, recipient, amount, attachment='', txFee=DEFAULT_TX_FEE, feeScale=DEFAULT_FEE_SCALE, timestamp=0):
        self._check_parameter(recipient=recipient, amount=amount, attachment=attachment, txFee=txFee, feeScale=feeScale)
        if timestamp == 0:
            timestamp = int(time.time() * 1000000000)
        sData = struct.pack(">B", PAYMENT_TX_TYPE) + \
                struct.pack(">Q", timestamp) + \
                struct.pack(">Q", amount) + \
                struct.pack(">Q", txFee) + \
                struct.pack(">H", feeScale) + \
                base58.b58decode(recipient) + \
                struct.pack(">H", len(attachment)) + \
                str2bytes(attachment)
        signature = bytes2str(sign(self.privateKey, sData))
        attachmentStr = bytes2str(base58.b58encode(str2bytes(attachment)))
        data = json.dumps({
            'senderPublicKey': self.publicKey,
            'recipient': recipient,
            'amount': amount,
            'fee': txFee,
            'feeScale': feeScale,
            'timestamp': timestamp,
            'attachment': attachmentStr,
            'signature': signature
        })
        return self.wrapper.request('/vsys/broadcast/payment', data)

    def lease(self, recipient, amount, txFee=DEFAULT_LEASE_FEE, feeScale=DEFAULT_FEE_SCALE, timestamp=0):
        self._check_parameter(recipient=recipient, amount=amount, txFee=txFee, feeScale=feeScale)
        data = self._form_lease_data(recipient=recipient, amount=amount, txFee=txFee, feeScale=feeScale, isLeasing=True)
        return self.wrapper.request('/leasing/broadcast/lease', data)

    def lease_cancel(self, recipient, amount, txFee=DEFAULT_CANCEL_LEASE_FEE, feeScale=DEFAULT_FEE_SCALE, timestamp=0):
        self._check_parameter(recipient=recipient, amount=amount, txFee=txFee, feeScale=feeScale)
        data = self._form_lease_data(recipient=recipient, amount=amount, txFee=txFee, feeScale=feeScale, isLeasing=False)
        return self.wrapper.request('/leasing/broadcast/cancel', data)

    def _form_lease_data(self, recipient, amount, txFee, feeScale, isLeasing):
        txType = None

        timestamp = int(time.time() * 1000000000)

        if isLeasing:
            txType = LEASE_TX_TYPE
        else:
            txType = LEASE_CANCEL_TX_TYPE

        sData = struct.pack(">B", txType) + \
                struct.pack(">Q", amount) + \
                struct.pack(">Q", txFee) + \
                struct.pack(">H", feeScale) + \
                struct.pack(">Q", timestamp)
        signature = bytes2str(sign(self.privateKey, sData))
        data = json.dumps({
            'senderPublicKey': self.publicKey,
            'recipient': recipient,
            'fee': txFee,
            'feeScale': feeScale,
            'timestamp': timestamp,
            'signature': signature
        })
        return data

    def _check_parameter(self, recipient, amount, txFee, feeScale, attachment=''):
        if not self.privateKey:
            msg = 'Private key required'
            throw_error(msg, MissingPrivateKeyException)
        if not self.chain.validate_address(recipient):
            msg = 'Invalid recipient address'
            throw_error(msg, InvalidAddressException)
        elif amount <= 0:
            msg = 'Amount must be > 0'
            throw_error(msg, InvalidParameterException)
        elif txFee < DEFAULT_TX_FEE:
            msg = f'Transaction fee must be >= {DEFAULT_TX_FEE}'
            throw_error(msg, InvalidParameterException)
        elif len(attachment) > MAX_ATTACHMENT_SIZE:
            msg = f'Attachment length must be <= {MAX_ATTACHMENT_SIZE}'
            throw_error(msg, InvalidParameterException)
        elif CHECK_FEE_SCALE and feeScale != DEFAULT_FEE_SCALE:
            msg = f'Wrong fee scale (currently, fee scale must be {DEFAULT_FEE_SCALE})'
            throw_error(msg, InvalidParameterException)
        elif self.balance() < amount + txFee:
            msg = f'Insufficient VSYS balance'
            throw_error(msg, InsufficientBalanceException)



import math
from rich.console import Console
from datetime import datetime

from market_place_cli.v_cloud_market_cli_common.service_display.display_common import utc_to_local

class MarketServiceRequest:

    def __init__(self, console: Console):
        self.console = console

    def get_category_name(self) -> str:
        msg = '[bright_green]Please enter category name: '
        return self.console.input(msg)

    def get_provider_name(self) -> str:
        msg = '[bright_green]Please enter provider name: '
        return self.console.input(msg)

    def get_service_id(self) -> str:
        msg = '[bright_green]Please enter service id: '
        return self.console.input(msg)

    def get_display_qr_code(self) -> bool:
        msg = '[bright_green]Show send payment QR code (default n) [Y/n]: '
        choice = self.console.input(msg)

        if choice.lower() not in ['y', 'n', '']:
            choice = 'n'
        if choice == '':
            choice = 'n'
        return choice.lower() == 'y'

    def user_choose_options(self, options) -> dict:
        result = {}
        for key in options:
            self.console.print(f'Choose service option for [bold magenta]{key}[/]:')
            choice_output = ''
            for index in range(len(options[key])):
                choice_output += f'[purple]{index + 1}[/]' + ' -- ' + options[key][index] + '\n'
            self.console.print(choice_output)
            choice = self._get_int_num('[bright_green]Please choose a number: ')
            while choice < 1 or choice > len(options[key]):
                choice = self._get_int_num('[bright_green]Please choose a number: ')
            result[key] = options[key][choice-1]
        return result

    def user_choose_duration(self, price_set):
        duration_result = {}
        self.console.print('\nPlease choose time duration ( HOUR ) on the left. '
                           'The right side is corresponding price factor.')
        self.console.print('[magenta]0: Enter A Custom Expiration Date in YYYY-MM-DD hh:mm:ss')
        for key in price_set['duration']:
            self.console.print('[magenta]' + str(key) + '[/] : ' + '[cyan]' + str(price_set['duration'][key]))
        self.console.print('\n')

        min_hour = self._get_min(price_set['duration'].keys())
        duration = self._get_int_num('[bright_green]Please enter the time duration number: ')
        if duration == 0:
            FMT = '%Y-%m-%d %H:%M:%S'
            date_str = self.console.input('[bright_green]Please enter expired date: ')
            time_delta = datetime.strptime(date_str, FMT) - datetime.now()
            total_time = math.ceil(time_delta.total_seconds() / 3600)
            if total_time <= 0:
                self.console.print('[bright_red]Invalid Expiration Date!!')
                return self.user_choose_duration(price_set)
            duration_result['time'] = total_time
            duration_result['expiredDate'] = utc_to_local(datetime.strptime(date_str, FMT)).strftime('%Y-%m-%d')
            return duration_result
        while duration < min_hour:
            self.console.print('[bright_red]The number you entered is not available.\n')
            duration = self._get_int_num('[bright_green]Please enter the time duration number: ')
        duration_result['time'] = duration
        return duration_result

    def _get_int_num(self, msg):
        try:
            choice = int(self.console.input(msg))
            return choice
        except ValueError:
            self.console.print('[bright_red]The input you entered is invalid.')
            return 0

    def _get_min(self, strList) -> int:
        result = []
        for a in strList:
            result.append(int(a))
        return min(result)



from __future__ import absolute_import

VERSION = '0.1.0'

def get_version():
    return VERSION

# import logging
#
# logging.getLogger('requests').setLevel(logging.WARNING)
# console = logging.StreamHandler
# console.setLevel(logging.ERROR)
# formatter = logging.Formatter('[%(levelname)s] %(message)s')
# console.setFormatter(formatter)
# logging.getLogger('').addHandler(console)


from .setting import *
from .api_wrapper import Wrapper


def create_api_wrapper(nodeHost=DEFAULT_NODE, apiKey=DEFAULT_NODE_API_KEY):
    return Wrapper(nodeHost, apiKey)


from .chain import *


def testnet_chain(apiWrapper=create_api_wrapper(DEFAULT_TESTNET_NODE, DEFAULT_TESTNET_API_KEY)):
    return Chain(TESTNET_CHAIN, TESTNET_CHAIN_ID, ADDRESS_VERSION, apiWrapper)


def mainnet_chain(apiWrapper=create_api_wrapper()):
    return Chain(DEFAULT_CHAIN, DEFAULT_CHAIN_ID, ADDRESS_VERSION, apiWrapper)



from .market_service_logic import MarketServiceLogic
from .order_service_logic import OrderServiceLogic
from .wallet_service_logic import WalletServiceLogic
from .user_service_logic import UserServiceLogic
from .initialization_logic import InitializationLogic


class ServiceLogicContainer:

    def __init__(self):
        self.container = {}

    def register(self, serviceLogic):
        self.container[serviceLogic.Name] = serviceLogic


il = InitializationLogic()
ml = MarketServiceLogic()
ol = OrderServiceLogic()
wl = WalletServiceLogic()
ul = UserServiceLogic()

Logics = ServiceLogicContainer()
Logics.register(il)
Logics.register(ml)
Logics.register(ol)
Logics.register(wl)
Logics.register(ul)
